from lxml import etree
from cssselect import GenericTranslator, SelectorError


class Locator:

    minimal_xml = '<!DOCTYPE _[<!ELEMENT _ EMPTY>]><_/>'

    @classmethod
    def contains(cls, element, *text):
        """
        Находит элемент, содержащий текст.
        Можно передать CSS или XPath, однако они будут превращены в XPath.
        """
        condition = ""
        for string in text:
            condition += "[contains(., {})]".format(
                GenericTranslator().xpath_literal(string)
            )

        xpath = '{0}{1}'.format(
            cls.to_xpath(element),
            condition
        )
        return xpath

    @classmethod
    def equal(cls, element, text):
        xpath = '{0}[{1}]'.format(
            cls.to_xpath(element),
            ".= %s" % GenericTranslator().xpath_literal(text)
        )
        return xpath

    @classmethod
    def element_at(cls, locator, position):
        """
        Элемент по индексу.
        Если передать отрицательное значение, то счёт пойдёт с конца.
        Первый элемент имеет индекс 1.
        """
        if isinstance(position, int) and position < 0:
            position += 1
            position = 'last()-{0}'.format(abs(position))

        if position == 0:
            raise ValueError('0 is not valid element position. '
                             'XPath expects first element to have index 1')
        return "({0})[position()={1}]".format(cls.to_xpath(locator), position)

    @classmethod
    def to_xpath(cls, selector):
        """
        Конвертирует CSS селектор в XPath.
        Если передать валидный XPath, то вернёт его без изменений.
        """
        try:
            return GenericTranslator().css_to_xpath(selector)
        except SelectorError:
            if cls.is_xpath(selector):
                return selector
        return None

    @classmethod
    def is_xpath(cls, selector):
        """
        Проверяет, что переданный слектор - XPath
        """
        doc = etree.XML(cls.minimal_xml)
        try:
            doc.xpath(selector)
        except etree.XPathEvalError:
            return False
        return True

    @classmethod
    def is_css(cls, selector):
        """
        Проверяет, что переданный слектор - CSS
        """
        try:
            GenericTranslator().css_to_xpath(selector)
        except SelectorError:
            return False
        return True



import time
from selenium.common.exceptions import NoSuchElementException
from selenium.common.exceptions import TimeoutException


POLL_FREQUENCY = 0.5  # How long to sleep inbetween calls to the method
IGNORED_EXCEPTIONS = (NoSuchElementException,)  # exceptions ignored during calls to the method


class Wait(object):
    def __init__(self, browser, timeout, poll_frequency=POLL_FREQUENCY, ignored_exceptions=None):
        self._browser = browser
        self._timeout = float(timeout)
        self._poll = poll_frequency
        if self._poll <= 0:
            self._poll = POLL_FREQUENCY
        exceptions = list(IGNORED_EXCEPTIONS)
        if ignored_exceptions is not None:
            try:
                exceptions.extend(iter(ignored_exceptions))
            except TypeError:  # ignored_exceptions is not iterable
                exceptions.append(ignored_exceptions)
        self._ignored_exceptions = tuple(exceptions)

    def __repr__(self):
        return '<{0.__module__}.{0.__name__} (session="{1}")>'.format(
            type(self), self._browser._driver_instance.session_id)

    def until(self, method, message=''):
        screen = None
        stacktrace = None

        end_time = time.time() + self._timeout
        while True:
            try:
                value = method(self._browser)
                if value:
                    return value
            except self._ignored_exceptions as exc:
                screen = getattr(exc, 'screen', None)
                stacktrace = getattr(exc, 'stacktrace', None)
            time.sleep(self._poll)
            if time.time() > end_time:
                break
        raise TimeoutException(message, screen, stacktrace)

    def until_not(self, method, message=''):
        end_time = time.time() + self._timeout
        while True:
            try:
                value = method(self._browser)
                if not value:
                    return value
            except self._ignored_exceptions:
                return True
            time.sleep(self._poll)
            if time.time() > end_time:
                break
        raise TimeoutException(message)



from selenium.common.exceptions import NoSuchElementException
from selenium.common.exceptions import StaleElementReferenceException


class visibility_of(object):
    def __init__(self, element):
        self.element = element

    def __call__(self, browser):
        try:
            return len(browser._match_visible(self.element)) > 0
        except StaleElementReferenceException:
            return False


class invisibility_of(object):
    def __init__(self, element):
        self.element = element

    def __call__(self, browser):
        try:
            return len(browser._match_visible(self.element)) == 0
        except (NoSuchElementException, StaleElementReferenceException):
            return True


class staleness_of(object):
    def __init__(self, element):
        self.element = element

    def __call__(self, browser):
        try:
            # Calling any method forces a staleness check
            el.is_enabled()
            return False
        except StaleElementReferenceException:
            return True



import yaml
import re
import os
from wasd.core import session
from wasd.common import LOGGER


class SettingsManager:

    _default_config_file = session.root_dir.joinpath('_wasd_settings.yml')

    _default_config = {
        'protocol': 'http',
        'host':     'localhost',
        'port':     '4444',
        'path':     '/wd/hub',
        'implicit_timeout': 5,
        'window_size':      '800x600'
    }

    @classmethod
    def init(cls, env):
        cls._load_env_file(env)
        cls._parse_window_size()

    @classmethod
    def _load_env_file(cls, env=None):
        if env is not None:
            env_file = session.env_dir.joinpath(f'{env}.yml')
        else:
            env_file = cls._default_config_file

        try:
            with open(env_file, 'r') as f:
                data = f.read()
        except IOError:
            LOGGER.error(f"Env file '{env_file}' not found.")
            raise

        cls._config = {
            **cls._default_config,
            **cls._prepare_config(yaml.safe_load(data))
        }

    @classmethod
    def get(cls, k):
        return cls._config[k]

    @classmethod
    def _prepare_config(cls, d):
        for k, v in d.items():
            if isinstance(v, dict):
                cls._prepare_config(v)
            else:
                if isinstance(v, list):
                    for i, _ in enumerate(v, start=0):
                        if isinstance(_, str):
                            d[k][i] = re.sub(
                                r'%(\w+)%',
                                lambda match: os.environ[match.group(1)], v[i]
                            )
                else:
                    if isinstance(v, str):
                        d[k] = re.sub(
                            r'%(\w+)%',
                            lambda match: os.environ[match.group(1)], v
                        )
        return d

    @classmethod
    def _parse_window_size(cls):
        raw_data = cls.get('window_size')
        if raw_data == 'maximize':
            return
        parsed = tuple(list(map(lambda v: int(v), raw_data.split('x'))))
        cls._config['window_size'] = parsed



# -*- coding: utf-8 -*-
from selenium.webdriver.common.by import By
from wasd.util.locator import Locator
from wasd.common.exceptions import MalformedLocatorException


# https://github.com/mfalesni/selenium-smart-locator

class Element:
    BY_MAPPING = {
        'partial_link_text':    By.PARTIAL_LINK_TEXT,
        'css':                  By.CSS_SELECTOR,
        'class_name':           By.CLASS_NAME,
        'link_text':            By.LINK_TEXT,
        'tag':                  By.TAG_NAME,
        'name':                 By.NAME,
        'xpath':                By.XPATH,
        'id':                   By.ID
    }
    REVERSE_BY_MAPPING = {v: k for k, v in BY_MAPPING.items()}

    def _get_by(self, by):
        try:
            if by in self.REVERSE_BY_MAPPING:
                by_ = self.REVERSE_BY_MAPPING[by]
            else:
                by_ = by
            return self.BY_MAPPING[by_]
        except KeyError:
            raise ValueError(f"'{by}' is not a recognized resolution strategy")


    def __init__(self, *args):
        ctx = None
        if len(args) == 1:
            if isinstance(args[0], Element):
                by = args[0].by
                val = args[0].val
                ctx = args[0].ctx
            elif isinstance(args[0], tuple):
                by = args[0][0]
                val = args[0][1]
            elif isinstance(args[0], str):
                by = By.CSS_SELECTOR if Locator.is_css(args[0]) else By.XPATH
                val = args[0]
        elif len(args) == 2:
            if isinstance(args[0], Element):
                by = args[0].by
                val = args[0].val
                ctx = args[1]
            elif isinstance(args[0], str):
                by = By.CSS_SELECTOR if Locator.is_css(args[0]) else By.XPATH
                val, ctx = args[0], args[1]
            elif isinstance(args[0], tuple):
                by = args[0][0]
                val = args[0][1]
                ctx = args[1]
        else:
            raise MalformedLocatorException(f"Wrong parameters specified for locator: {args}")


        self.by = self._get_by(by)
        self.val = val
        self.ctx = ctx


    def locator(self):
        '''Возвращает кортеж (By, Value)'''
        return (self.by, self.val)


    def __str__(self):
        return "({0} = '{1}', ctx = '{2}')".format(self.by, self.val, self.ctx)


class ShadowElement(Element):

    def __str__(self):
        return "(ShadowRoot::{0} = '{1}', ctx = '{2}')".format(self.by, self.val, self.ctx)



from wasd.core import SettingsManager
import time
from selenium.webdriver.support.events import AbstractEventListener


class ElementHighlightListener(AbstractEventListener):
    '''
    Подсвечивает элемент с которым работает
    '''
    
    inset_shadow = 'box-shadow: inset 0 0 5px 1px #0a0;'


    def after_find(self, by, value, driver):
        self.highlight(by, value, driver)


    def highlight(self, by, value, driver):
        if by == 'css selector':
            elem = self.js_find_by_css(driver, value)
        if by == 'xpath':
            elem = self.js_find_by_xpath(driver, value)

        original_style = self.get_style(driver, elem)
        concat_style = f"{original_style};{self.inset_shadow}"

        self.apply_style(driver, elem, concat_style)
        time.sleep(.2)
        self.apply_style(driver, elem, original_style)


    def get_style(self, driver, web_element):
        get_style_script = '''
            if (arguments[0]) {
                return arguments[0].getAttribute('style');
            } else {
                return "";
            }'''


        return driver.execute_script(get_style_script, web_element)


    def apply_style(self, driver, web_element, style):
        apply_style_script = '''
            if (arguments[0]) {{
                arguments[0].setAttribute('style', '{0}');
            }}'''.format(style)

        return driver.execute_script(apply_style_script, web_element)


    def js_find_by_css(self, driver, selector):
        script = f'''return document.querySelector({selector.__repr__()});'''
        return driver.execute_script(script)


    def js_find_by_xpath(self, driver, selector):
        script = f'''
        return document.evaluate({selector.__repr__()}, document, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
        '''
        return driver.execute_script(script)



import time
import re
import os
import uuid
import json
from hamcrest import *
from pathlib import Path
from urllib.parse import urljoin, urlparse
from contextlib import contextmanager
from selenium import webdriver
from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.events import EventFiringWebDriver
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.remote.file_detector import LocalFileDetector
from wasd.core import session
from wasd.core import SettingsManager
from wasd.common import log_step
from wasd.wd.element import Element, ShadowElement
from wasd.wd.listener import ElementHighlightListener

from wasd.support.wait import Wait
from wasd.support.expected_conditions import visibility_of, invisibility_of


class Browser:

    def __init__(self):
        self.new_driver()

    _driver_instance = None
    _session_snapshots = {}

    def new_driver(self):
        command_executor = "{0}://{1}:{2}{3}".format(
            SettingsManager._config['protocol'],
            SettingsManager.get('host'),
            SettingsManager.get('port'),
            SettingsManager.get('path')
        )

        remote_driver = webdriver.Remote(
            command_executor=command_executor,
            desired_capabilities=SettingsManager.get('capabilities')
        )

        if session.use_listener:
            self._driver_instance = EventFiringWebDriver(remote_driver, ElementHighlightListener())
        else:
            self._driver_instance = remote_driver

        self._driver_instance.implicitly_wait(SettingsManager.get('implicit_timeout'))
        self._driver_instance.maximize_window()

        return self._driver_instance

    def get_driver(self):
        if not self._driver_instance:
            self.new_driver()
        return self._driver_instance

    def close_driver(self):
        if self._driver_instance:
            self._driver_instance.quit()
            self._driver_instance = None

    def attach_file(self, element, file_name):
        log_step(f"Attach file {file_name}")
        file_path = session.data_dir.joinpath(file_name)

        if not Path.exists(file_path):
            raise IOError(f"File does not exist: {file_path}")

        self.get_driver().file_detector = LocalFileDetector()
        self.fill_field(element, str(file_path))

    def open(self, path):
        """
        Переходит по относительному URL (/login, /admin/users, etc ...)

        Args:
            path (str): Путь

        Examples:
            >>> browser.open("/")
            # На главную

            >>> browser.open("/admin")
            # В админку
        """
        url = urljoin(SettingsManager.get('url'), path)
        log_step(f"Open {url}")
        self.get_driver().get(url)

    def open_url(self, url):
        """
        Переходит по абсолютному URL.

        Args:
            url (str): URL

        Examples:
            >>> browser.open_url("https://google.com")
        """
        log_step(f"Open URL {url}")
        self.get_driver().get(url)

    def refresh(self):
        """
        Обновляет страницу
        """
        log_step("Refresh page")
        self.get_driver().refresh()

    def grab_console_log(self):
        '''
        Получить лог браузерной консоли. Буффер очищается после каждого запроса.
        Требует сapability ``loggingPrefs: { browser: 'INFO' }``.

        Returns:
            list[str]: список строк
        '''
        log_step("Grab browser console log")
        return self.get_driver().get_log('browser')

    def grab_page_html(self):
        """
        Возвращает html всей страницы.

        Returns:
            str: html
        """
        log_step("Grab page html")
        return self.get_driver().page_source

    def grab_html_from(self, element):
        """
        Возвращает html элемента.

        Returns:
            str: html
        """
        log_step(f"Grab html from {element}")
        el = self._match_first_or_fail(element)
        return el.get_attribute('outerHTML')

    def clear_field(self, input_element):
        """
        Очищает поле.

        Args:
            input_element (Element): Поле ввода

        Examples:
            >>> browser.clear_field(Element("#input"))
        """
        log_step(f"Clear field {input_element}")
        el = self._match_first_or_fail(input_element)
        el.clear()

    def fill_field(self, element, text):
        """
        Заполняет текстом поле ввода

        Args:
            element (Element): Поле ввода
            text    (str):  Текст

        Examples:
            >>> browser.fill_field(Element("//input[@type='text']"), "Hello World!")
        """
        log_step(f"Fill field {element} with '{text}'")
        el = self._match_first_or_fail(element)
        el.clear()
        el.send_keys(text)

    def fill_field_with_delay(self, element, text, delay=0.1):
        """
        Заполняет текстом поле ввода делая паузу между каждым символом.

        Args:
            element (Element): Поле ввода
            text    (str):  Текст
            delay   (float, optional): Задержка (сек.), по умолчанию 0.1

        Examples:
            >>> browser.fill_field(Element("//input[@type='text']"), "Hello World!", 0.2)
        """
        log_step(f"Fill field {element} with '{text}' & delay '{delay}'")
        field = self._match_first_or_fail(element)
        for ch in text:
            field.send_keys(ch)
            self.sleep(delay)

    def press_key(self, element, *chars):
        """
        Нажимает кнопки.
        Чтобы ввести символы с модификатором (ctrl, alt, shift, meta), передай кортеж
        у которого первый элемент модификатор, а второй текст.

        Args:
            element (Element):      Элемент, куда посылать нажатия
            *chars  (str|tuple):    Что вводить. Может быть строкой или кортежем в модификатором

        Examples:
            Дано: поле, в которое введён текст ``old``:\n
            ``<input id="input" value="old" />``

            >>> from selenium.webdriver.common.keys import Keys
            >>> field = Find("#input")
            >>> browser.press_key(field, "A")
            oldA
            >>> browser.press_key(field, ("ctrl", "a"), "new")
            new
            >>> browser.press_key(field, ("shift", "333"), "2", "x")
            old###2x
            >>> browser.press_key(field, Keys.BACKSPACE)
            ol
        """
        log_step(f"Press keys {chars} on {element}")
        el = self._match_first_or_fail(element)
        for char in chars:
            el.send_keys(*self._convert_key_modifier(char))

    def append_field(self, element, text):
        """
        Присоединяет переданный текст к элементу.

        Args:
            element     (Element): элемент
            text        (str): текст

        Examples:
            >>> browser.append_field(Find('#field'), 'foo bar baz')
        """
        log_step(f"Append field {element}, '{text}'")
        field = self._match_first_or_fail(element)
        field.send_keys(text)

    def wait_for_element_visible(self, element, timeout=5, poll_frequency=0.5):
        """
        Ждёт до ``timeout`` сек. видимости элемента.
        Если элемент не появился, бросает timeout exception.

        Args:
            element (Element): Элемент, который ждём
            timeout (int, optional): Таймаут ожидания (сек.), по умолчанию 5.

        Raises:
            TimeoutException

        Examples:
            >>> browser.wait_for_element_visible(Element("#header", 15))

        """
        log_step(f"Wait for element visible {element}")
        self.wait(timeout, poll_frequency).until(visibility_of(element))

    def wait_for_element_not_visible(self, element, timeout=5, poll_frequency=0.5):
        """
        Ждёт до ``timeout`` сек. исчезновения элемента.
        Если элемент остался видимым, бросает timeout exception.

        Args:
            element (Element): Элемент, который ждём
            timeout (int, optional): Таймаут ожидания (сек.), по умолчанию 5.

        Raises:
            TimeoutException

        Examples:
            >>> browser.wait_for_element_not_visible(Element("#header", 15))
        """
        log_step(f"Wait for element not visible {element}")
        self.wait(timeout, poll_frequency).until(invisibility_of(element))

    def see_element(self, element, attributes={}):
        """
        Проверяет, что элемент существует и видим.
        Также можно указать атрибуты этого элемента.

        Args:
            element (Element): Элемент
            attributes (dict, optional): Словарь атрибутов

        Examples:
            >>> browser.see_element(Element("tr"))
            >>> browser.see_element(Element("tr"), {"some-attr": "some-value"})
        """
        log_step(f"See element {element}, {attributes}")
        self._enable_implicit_wait()
        els = self._match_visible(element)
        self._disable_implicit_wait()
        els = self._filter_by_attributes(els, attributes)
        assert_that(els, is_not(empty()))

    def see_text(self, text, element=None):
        """
        Проверяет, что страница содержит текст (с учётом регистра).
        Можно передать элемент, чтобы искать текст только в нём.

        Args:
            text (str): Искомый текст
            element (Element, optional): Элемент

        Examples:
            >>> browser.see_text("Администрирование")
            >>> browser.see_text("Выйти", Element("h1"))
        """
        log_step(f"See text {text} in {element}")
        self._enable_implicit_wait()
        text_from_page = self.grab_visible_text(element)
        self._disable_implicit_wait()
        assert_that(text_from_page, contains_string(text))

    def see_in_field(self, input_element, needle):
        """
        Проверяет, что текст поля ввода равен переданному значению.

        Args:
            input_element (Element): Элемент
            needle (str): Текст

        Examples:
            >>> browser.see_in_field(Element("input#foo"), "Hello World!")
        """
        log_step(f"See text '{needle}' in field {input_element}")
        val = self.grab_value_from(input_element)
        assert_that(val, equal_to(needle))

    def see_number_of_elements(self, element, expected):
        """
        Проверяет, что на странице определённое кол-во элементов.

        Args:
            element (Element): Элемент
            expected (int|tuple): Ожидаемое кол-во

        Returns:
            None

        Raises:
            AssertionError

        Examples:
            >>> browser.see_number_of_elements( E('tr'), 10 )
            >>> browser.see_number_of_elements( E('tr'), (4, 6) ) # Между 4 и 6, включительно
        """
        log_step(f"See number of elements {element}, {expected}")
        count = len(self._match_visible(element))
        if isinstance(expected, tuple):
            a, b = expected
            assert_that(a <= count <= b, is_(True),
                        f'Number of elements expected to be in range {expected}, but was {count}')
        else:
            assert_that(count, equal_to(expected),
                        f'Number of elements expected to be {expected}, but was {count}')

    def grab_visible_text(self, element=None):
        """
        Получает видимый текст всей страницы или элемента.

        Args:
            element (Element, optional): Элемент

        Returns:
            str: Текст

        Examples:
            >>> browser.grab_visible_text()
            >>> browser.grab_visible_text(Element("h1"))
        """
        log_step(f"Get visible text from {element}")
        if element is not None:
            return self.grab_text_from(element)

        els = self._match(Element("body"))
        if len(els) == 0:
            return ''

        return els[0].text

    def click(self, element):
        """
        Кликает по элементу.

        Args:
            element (Element): Элемент

        Examples:
            >>> browser.click(Element("#logout"))
        """
        log_step(f"Click {element}")
        self._match_first_or_fail(element).click()

    def double_click(self, element):
        """
        Дважды кликает по элементу.

        Args:
            element (Element): Элемент

        Examples:
            >>> browser.double_click(Element("#login"))
        """
        log_step(f"Double click {element}")
        target = self._match_first_or_fail(element)
        ActionChains(self._driver_instance).double_click(target).perform()

    def js_click(self, element):
        """
        Click via JS

        Args:
            element (Element)
        """
        log_step(f"Click via JS {element}")
        script = "var rect = arguments[0].getBoundingClientRect();" \
                 "arguments[0].dispatchEvent(new MouseEvent('click', {" \
                 "  'bubbles': true," \
                 "  'cancelable': true," \
                 "  'view': window," \
                 "  'clientX': rect.left + rect.width/2," \
                 "  'clientY': rect.top + rect.height/2" \
                 "}));"
        self.execute_js(script, self._match_first_or_fail(element))

    def grab_visible(self, element):
        """
        Получает видимые элементы.

        Args:
            element (Element, optional): Элемент

        Returns:
            selenium.webdriver.remote.webelement.WebElement[]: Список элементов

        Examples:
            >>> browser.grab_visible(Element("th"))
        """
        log_step(f"Grab visible {element}")
        return self._match_visible(element)

    def grab_text_from(self, element=None):
        """
        Получает текст из переданного элемента
        или из body, если передан None.

        Args:
            element (Element, optional): Элемент

        Returns:
            str: Текст

        Examples:
            >>> browser.grab_text_from(Element("h1"))
        """
        log_step(f"Get text from {element}")
        if element is None:
            element = Element("body")

        el = self._match_first_or_fail(element)
        return el.text

    def js_grab_text_from(self, element):
        log_step(f"Grab text via JS from {element}")
        return self.execute_js("return arguments[0].textContent;", self._match_first_or_fail(element))

    def grab_attribute_from(self, element, attribute):
        """
        Получает значение атрибута.

        Args:
            element     (Element): Элемент
            attribute   (str):  Атрибут

        Returns:
            str: Значение атрибута

        Examples:
            ``<span id="tooltip" title="hello_world"> </span>``

            >>> browser.grab_attribute_from(Element('#tooltip'), 'title')
            "hello_world"
        """
        log_step(f"Grab attribute from {element}")
        el = self._match_first_or_fail(element)
        return el.get_attribute(attribute)

    def grab_value_from(self, input_element):
        """
        Получает value атрибут из поля.
        Используется, чтобы получить введённый текст.

        Args:
            element (Element): Элемент

        Returns:
            str: Значение value атрибута

        Examples:
            ``<input id="login" value="ivanov">``

            >>> browser.grab_value_from(Element('#login'))
            "ivanov"
        """
        log_step(f"Grab value from {input_element}")
        el = self._match_first_or_fail(input_element)
        return el.get_attribute('value')

    def grab_multiple(self, elements):
        """
        Получает тексты подходящих элементов и возващает их как список.

        Args:
            element (Element): Элемент

        Returns:
            str[]: Список текстов

        Examples:
            .. code-block:: html

                <tr> foo </tr>
                <tr> bar </tr>
                <tr> baz </tr>


            >>> browser.grab_multiple( Element('tr') )
            ["foo", "bar", "qaz"]
        """
        log_step(f"Grab multiple {elements}")
        els = self._match(elements)
        return list(map(lambda el: el.text, els))

    def js_grab_multiple(self, element):
        log_step(f"Grab multiple via JS from {element}")
        script = "return Object.values(arguments).map(function(e) { return e.textContent.trim(); });"
        return self.execute_js(script, *self._match(element))

    def save_screenshot(self, name=None):
        """
        Делает скриншот видимой части страницы и сохраняет в ``_output/debug``.

        Args:
            name (str, optional): Имя файла

        Returns:
            pathlib.Path: Путь до файла
        """
        debug_dir = session.output_dir.joinpath('debug')
        if not os.path.exists(debug_dir):
            os.makedirs(debug_dir, exist_ok=True)

        if name is None:
            name = uuid.uuid4()

        file_path = debug_dir.joinpath(f"{name}.png")
        log_step(f"Save screenshot to {file_path}")
        self.get_driver().get_screenshot_as_file(str(file_path))
        return file_path

    def get_screenshot_binary(self):
        """
        Возвращает скриншот в бинарном формате.

        Returns:
            bytes
        """
        log_step(f"Get screenshot as png")
        return self.get_driver().get_screenshot_as_png()

    def move_mouse_over(self, element):
        """
        Передвигает курсор в центр элемента

        Args:
            element (Element): элемент
        """
        log_step(f"Move mouse over {element}")
        el = self._match_first_or_fail(element)
        ActionChains(self.get_driver()).move_to_element(el).perform()

    def switch_to_iframe(self, frame=None):
        """
        Переключяет контекст в iframe.
        Чтобы перейти в родительский фрейм - вызывается без параметров.

        Args:
            frame (Element): iframe

        Examples:
            ``<iframe name="another_frame" src="http://example.com">``

            >>> frame = Element("frame[name = 'another_frame']")
            >>> browser.switch_to_iframe(frame)
            >>> browser.switch_to_iframe() # в родительский фрейм
        """
        log_step(f"Switch to iframe {frame}")
        if frame is not None:
            el = self._match_first_or_fail(frame)
            self.get_driver().switch_to.frame(el)
        else:
            self.get_driver().switch_to.parent_frame()

    def close_window(self):
        """
        Закрывает текущее окно и переключается в предыдущее.

        Returns:
            window handle
        """
        log_step("Close window")
        prev_window = self.get_relative_window_handle(-1)
        self.get_driver().close()
        self.get_driver().switch_to.window(prev_window)
        return prev_window

    def switch_to_next_window(self, offset=1):
        """
        Переключается в следующее окно.

        Args:
            offset (int, optional): По умолчанию 1

        Returns:
            window handle
        """
        log_step("Switching to next window")
        next_window = self.get_relative_window_handle(offset)
        self.get_driver().switch_to.window(next_window)
        return next_window

    def switch_to_prev_window(self, offset=1):
        """
        Переключается в предыдущее окно.

        Args:
            offset (int, optional): По умолчанию 1

        Returns:
            window handle
        """
        log_step("Switching to prev window")
        next_window = self.get_relative_window_handle(0-offset)
        self.get_driver().switch_to.window(next_window)
        return next_window

    def new_window(self, url="about:blank"):
        """
        Открывает новое окно и переключается в него.

        Args:
            url (str, optional): По умолчанию 'about:blank'

        Returns:
            window handle
        """
        log_step(f"Open new window with url '{url}''")
        original_handles = self._driver_instance.window_handles
        self.execute_js(f"window.open('{url}')")
        new_handles = self._driver_instance.window_handles
        new_window_hanle = (set(new_handles) - set(original_handles)).pop()
        self._driver_instance.switch_to.window(new_window_hanle)
        return new_window_hanle

    @contextmanager
    def in_frame(self, frame, exit_to='parent'):
        log_step(f"In frame {frame}")

        if exit_to not in ['parent', 'default']:
            raise TypeError(f"'exit_to' must be 'parent' or 'default', '{exit_to}' given")

        switch = {
            'parent': self.get_driver().switch_to.parent_frame,
            'default': self.get_driver().switch_to.default_content
        }
        try:
            self.switch_to_iframe(frame)
            yield
        finally:
            switch[exit_to]()

    def save_session_snapshot(self, name):
        """
        Сохраняет куки чтобы загрузить их в другом тесте.
        Например, если в каждом тесте необходима авторизация,
        можно выполнить этот сценарий единожды,
        а в других тестах только восстанавливать сессию.

        Args:
            name (str): Название сесиии

        Examples:
            >>> def login():
            >>>     # Если сессия существует - пропустить логин
            >>>     if browser.load_session_snapshot("login"):
            >>>         return
            >>>     browser.fill_field( Element("#username") )
            >>>     browser.fill_field( Element("#password") )
            >>>     browser.click( Element("#submit") )
            >>>     browser.save_session_snapshot("login")
        """
        log_step(f"Save session snapshot '{name}'")
        self._session_snapshots[name] = self.get_driver().get_cookies()

    def set_style(self, element, style, value, is_important=True):
        log_step(f"Set style {style}={value} on {element}")
        if is_important:
            script = f"arguments[0].style.setProperty('{style}', '{value}', 'important')"
        else:
            script = f"arguments[0].style.setProperty('{style}', '{value}')"
        self.execute_js(script, self._match_first_or_fail(element))

    def load_session_snapshot(self, name):
        """
        Загружает сохранённую сессию.
        Смотри :func:`~wd.browser.Browser.save_session_snapshot`.

        Args:
            name (str): Имя сессии

        Returns:
            bool
        """
        log_step(f"Load session snapshot '{name}'")
        if name not in self._session_snapshots:
            return False

        for cookie in self._session_snapshots[name]:
            self.set_cookie(cookie["name"], cookie["value"], cookie)

        return True

    def set_cookie(self, name, value, params={}):
        """
        Задаёт куку с именем и значением.
        Можно передать дополнительные параметры.

        Args:
            name    (str): Имя куки
            value   (str): Значение куки
            params  (dict, optional):
                Доп. параметры:

                    - domain
                    - path
                    - expiry
                    - secure
                    - httpOnly
        """
        log_step(f"Set cookie '{name}': '{value}'")
        params["name"] = name
        params["value"] = value

        if "domain" not in params:
            url_parts = urlparse(SettingsManager.get("url"))
            if "netloc" in url_parts:
                params["domain"] = url_parts["netloc"]

        defaults = {
            "path": '/',
            "expiry": int(time.time()) + 86400,
            "secure": False,
            "httpOnly": False,
        }

        for k, v in defaults.items():
            if k not in params:
                params[k] = v

        self.get_driver().add_cookie(params)

    def scroll_to(self, element, offset_x=0, offset_y=0):
        """
        Скроллит к центру переданного элемента.
        Дополнительное смещение, расчитываемое от левого верхнего угла элемента,
        можно задать параметрами `offset_x` и `offset_y` (в пикселях).

        Важно: вызывает метод scrollTo на объекте window.
        Если элемент находится в каком то контейнере с `overflow: scoll`,
        то данный метод не сработает, см. :func:`~wd.browser.Browser.scroll_into_view`.

        Args:
            element     (Element): элемент
            offset_x    (int, optional): смещение по X (px)
            offset_y    (int, optional): смещение по Y (px)
        """
        log_step(f"Scroll to {element}")
        el = self._match_first_or_fail(element)
        x = el.location['x'] + offset_x
        y = el.location['y'] + offset_y
        self.execute_js(f"window.scrollTo({x}, {y})")

    def scroll_into_view(self, element, scroll_into_view_options={"block": "center"}):
        """
        Скролит страницу к элементу

        Args:
            element (Element): Элемент
            scroll_into_view_options (dict, optional): Параметры, по умолчанию {'block': 'center'}
                https://developer.mozilla.org/ru/docs/Web/API/Element/scrollIntoView
        """
        log_step("Scroll into view")
        el = self._match_first_or_fail(element)
        self.execute_js(f"arguments[0].scrollIntoView({json.dumps(scroll_into_view_options)})", el)

    def delete_all_cookies(self):
        """
        Удаляет все куки.
        """
        log_step("Delete all cookies")
        self.get_driver().delete_all_cookies()

    def dont_see_element(self, element, attributes={}):
        """
        Проверяет, что элемент невидим.

        Args:
            element     (Element): Элемент
            attributes  (dict, optional): Атрибуты

        Raises:
            AssertionError
        """
        log_step(f"Dont see element {element}, {attributes}")
        els = self._match_visible(element)
        els = self._filter_by_attributes(els, attributes)
        assert_that(els, is_(empty()))

    def element_has_attribute(self, element, attr, expected_value=None):
        """
        Проверяет наличие атрибута у элемента.
        Опционально приинимает значение атрибута.

        Args:
            element         (Element): Элемент
            attr            (str): Имя атрибута
            expected_value  (string, optional): Ожидаемое значание

        Returns:
            bool

        Examples:
            ``<div class='foo bar bazqux'> </div>``

            >>> browser.element_has_attribute( Element('div'), 'class' )
            True
            >>> browser.element_has_attribute( Element('div'), 'class', 'foo' )
            True
            >>> browser.element_has_attribute( Element('div'), 'class', 'baz' )
            False
        """
        log_step(f"Element {element} has attribute {attr}: {expected_value}")

        is_attribue_presents = self.execute_js(
            f"return arguments[0].hasAttribute('{attr}');",
            self._match_first_or_fail(element)
        )

        if not is_attribue_presents:
            return False

        if expected_value is None:
            return True
        else:
            actual_val = self.grab_attribute_from(element, attr)

            values_list = re.sub(r'\s+', ' ', actual_val).strip().split(' ')
            if expected_value in values_list:
                return True
            else:
                return False

        # actual_val = self.grab_attribute_from(element, attr)

        # if actual_val is None:
        #     return False
        # elif expected_value is None:
        #     return True
        # else:
        #     values_list = re.sub(r'\s+', ' ', actual_val).strip().split(' ')
        #     if expected_value in values_list:
        #         return True
        #     else:
        #         return False

    def execute_js(self, script, *args):
        """
        Выполняет js на странице.

        Args:
            script  (str): JS скрипт
            *args   (obj): Аргументы

        Returns:
            То, что возвращает скрипт.

        Examples:
            ``<span id="tooltip" data-attr="foo"> </span>``

            >>> script = 'return arguments[0].getAttribute("data-attr");'
            >>> browser.execute_js(script, browser.find(Element("#tooltip")))
            "foo"
        """
        log_step(f"Execute JS '{script}'")
        return self.get_driver().execute_script(script, *args)

    def sleep(self, secs):
        """
        Просто ждёт ``secs`` секунд.

        Args:
            secs (float): Сколько секунд спать

        Raises:
            Exception: Если передали больше 1000 сек.
        """
        log_step(f"Sleep {secs}")
        if secs >= 1000:
            raise Exception("Waiting for more then 1000 seconds: 16.6667 mins")
        time.sleep(secs)

    def wd_wait(self, timeout=10, poll_frequency=0.5):
        """
        Возвращает объект WebDriverWait.

        Args:
            timeout         (int, optional): Таймаут ожидания, по умолчанию 10
            poll_frequency  (float, optional): Частота опроса, по умолчанию 0.5

        Returns:
            WebDriverWait
        """
        return WebDriverWait(self.get_driver(), timeout, poll_frequency)

    def wait(self, timeout=10, poll_frequency=0.5):
        """
        Возвращает объект wasd.support.wait.Wait
        """
        return Wait(self, timeout, poll_frequency)

    def scroll_top(self):
        log_step("Window scroll top")
        self.execute_js("window.scrollTo(0, 0)")

    def get_relative_window_handle(self, offset):
        original_handles = self._driver_instance.window_handles
        this_window = self._driver_instance.current_window_handle
        idx = original_handles.index(this_window)
        next_window = original_handles[(idx+offset) % len(original_handles)]
        return next_window

    def find(self, element):
        context = self.get_driver()
        try:
            if element.ctx is not None:
                context = self.find(element.ctx)

            if isinstance(element, ShadowElement):
                el = context.find_element(*element.locator())
                return self.execute_js('return arguments[0].shadowRoot;', el)
            else:
                return context.find_element(*element.locator())
        except NoSuchElementException as e:
            e.msg = f"No such element: Unable to locate element: {element}"
            raise

    def finds(self, element):
        context = self.get_driver()
        if element.ctx is not None:
            context = self.find(element.ctx)

        return context.find_elements(*element.locator())

    def _filter_by_attributes(self, els, attributes):
        for k, v in attributes.items():
            els = list(filter(
                lambda el: el.get_attribute(k) == v,
                els))
        return els

    def _match_visible(self, element):
        els = self._match(element)
        nodes = list(filter(
            lambda el: el.is_displayed(),
            els))
        return nodes

    def _match_first_or_fail(self, element):
        self._enable_implicit_wait()
        els = self._match(element)
        self._disable_implicit_wait()
        if not els:
            raise Exception("Element {0} was not found.".format(element))
        return els[0]

    def _match(self, element):
        return self.finds(element)

    def _enable_implicit_wait(self):
        self.get_driver().implicitly_wait(SettingsManager.get('implicit_timeout'))

    def _disable_implicit_wait(self):
        self.get_driver().implicitly_wait(0)

    def _convert_key_modifier(self, keys):
        if not isinstance(keys, tuple):
            return (keys,)
        if keys[0] in ["ctrl", "contol"]:
            return (Keys.CONTROL, keys[1])
        if keys[0] == "alt":
            return (Keys.ALT, keys[1])
        if keys[0] == "shift":
            return (Keys.SHIFT, keys[1])
        if keys[0] == "meta":
            return (Keys.META, keys[1])
        return keys



import os
import sys
from pathlib import Path
from termcolor import cprint


def show_usage():
    print('Type "wasd scaffold" for generate project stucture at current directory')
    print('* (Use "pytest" for running tests) *\n')


def main():
    num_args = len(sys.argv)

    if num_args != 2 or sys.argv[1] != 'scaffold':
        show_usage()
        return

    proj_dir = Path.cwd()

    if os.path.exists(proj_dir.joinpath('_wasd_settings.yml')):
        cprint(f'Project under "{proj_dir}" already exists', 'red')
        return

    test_dir = proj_dir.joinpath('tests')
    output_dir = proj_dir.joinpath('_output')
    os.mkdir(test_dir)
    os.mkdir(output_dir)
    os.mkdir(proj_dir.joinpath('_data'))
    

    this_directory = Path(__file__).parent.absolute()

    # source : dist
    files_to_create = [
        ('config_browsers.json.txt',    ['config', 'browsers.json']),
        ('tests_conftest.py.txt',   ['tests', 'conftest.py']),
        ('tests_sample.py.txt',     ['tests', 'test_something.py']),
        ('env_stable.yml.txt',      ['_env', 'stable.yml']),
        ('_wasd_settings.yml.txt',   ['_wasd_settings.yml']),
        ('base_page.py.txt',    ['page', 'base_page.py']),
        ('home_page.py.txt',    ['page', 'home_page.py']),
        ('requirements.txt',    ['requirements.txt']),
        ('__init__.py.txt', ['tests', '__init__.py']),
        ('__init__.py.txt', ['page', '__init__.py']),
        ('pytest.ini.txt',  ['pytest.ini']),
        ('tasks.py.txt',    ['tasks.py'])
    ]

    for source, dist in files_to_create:
        source_path = this_directory.joinpath('boilerplate', source)
        dist_path = proj_dir.joinpath(*dist)
        dist_path.parent.mkdir(parents=True, exist_ok=True)

        with open(source_path, 'r') as sf:
            sd = sf.read()
            with open(dist_path, 'w+') as df:
                cprint(f"> {dist_path}", 'cyan')
                df.write(sd)
    cprint('  --- ', 'cyan')



import pytest
from pathlib import Path
from wasd.core import SettingsManager
from wasd.core import session
from termcolor import colored
from wasd.common.logger import LOGGER, __fake_logger
from wasd.wd import Browser
import uuid
import os
from datetime import datetime as dt


if not os.path.exists(Path.cwd().joinpath('_wasd_settings.yml')):
    '''
    Do not load plugin if there is no `_wasd_settings.yml` file
    '''
    pass

else:
    def pytest_addoption(parser):
        parser.addoption("--env", action="store")
        parser.addoption("--listener", action="store_const", const=False)
        parser.addoption("--save-screenshot", action="store_const", const=False)
        parser.addoption("--steps", action="store_const", const=False)

    def pytest_configure(config):
        config.addinivalue_line("markers", "want_to(arg): hello world.")

    @pytest.fixture(scope='session', autouse=True)
    def init_settings_fixture(request):
        session.env = request.config.getoption("--env")
        session.use_listener = True if request.config.getoption("--listener") is not None else False
        session.save_screenshot = True if request.config.getoption("--save-screenshot") is not None else False
        session.steps = True if request.config.getoption("--steps") is not None else False
        SettingsManager.init(session.env)

    def pytest_runtest_logfinish(nodeid, location):
        print()

    @pytest.hookimpl(tryfirst=True)
    def pytest_runtest_setup(item):
        __fake_logger.log(61, "\n" + item.pretty_id)
        if session.steps:
            __fake_logger.log(61, colored('Scenario --', 'yellow'))

    @pytest.hookimpl(tryfirst=True)
    def pytest_collection_modifyitems(config, items):
        for item in items:
            pretty_id = colored(item.parent.parent.name + ": ", 'magenta', attrs=['bold'])

            # Если будет несколько маркеров, то просто брать последний, не кидая ошибку
            want_to_list = list(filter(lambda m: m.name == 'want_to', item.own_markers))
            if len(want_to_list) >= 1:
                func_id = want_to_list[-1].args[0]
            else:
                splitted = item.name.split('_')
                captalized = splitted[0].capitalize() + ' ' + ' '.join(splitted[1:])
                func_id = captalized

            pretty_id += colored(func_id, 'white', attrs=['bold'])

            item.pretty_id = pretty_id

    @pytest.mark.hookwrapper
    def pytest_runtest_makereport(item, call):
        outcome = yield
        report = outcome.get_result()

        browser = item.funcargs.get("browser")
        if browser and isinstance(browser, Browser):
            driver = browser._driver_instance
            item.browser = browser

            if report.failed:
                if driver is None:  # If driver was closed
                    return
                item.screenshot_path, item.screenshot_binary = take_screenshot(driver, item)

    def take_screenshot(driver, item):
        now_timestamp = dt.now().isoformat(timespec="seconds")
        id_ = f"{item.location[2]}__{now_timestamp}.png"
        screenshot_path = str(session.output_dir.joinpath(id_))

        if session.save_screenshot:
            if not os.path.exists(session.output_dir):
                os.mkdir(session.output_dir)
            driver.get_screenshot_as_file(screenshot_path)
            __fake_logger.log(61, colored(f'◉ Screenshot saved to: {screenshot_path}', 'green'))

        screenshot_binary = driver.get_screenshot_as_png()
        return (screenshot_path, screenshot_binary)


def load():
    """Required for `pluggy` to load a plugin from setuptools entrypoints."""



from pathlib import Path
import os


root_dir = Path(os.getcwd())
test_dir = root_dir.joinpath('tests')
env_dir = root_dir.joinpath('_env')
output_dir = root_dir.joinpath('_output')
data_dir = root_dir.joinpath('_data')

env = None
use_listener = None
save_screenshot = None
steps = None


from setuptools import setup, find_packages
from pathlib import Path
import sys
import os


long_description = None
this_directory = Path(__file__).parent.absolute()

try:
    with open(this_directory.joinpath('README.md'), 'rb') as f:
        long_description = f.read().decode('utf-8')
except IOError:
    long_description = 'See https://github.com/Aurococcus/wasd'


if sys.argv[-1] == 'publish':
    print("\n*** Cleanup dist: ***")
    os.system('rm -rf dist/ build/')

    print("\n*** Creating new tar/wheel: ***\n")
    os.system('python setup.py sdist bdist_wheel')
    
    try:
        import twine
        print('\n*** twine OK ***\n')
    except ImportError:
        print('\n*** Installing twine: ***\n')
        os.system('python -m pip install twine')

    print("\n*** Publishing to PyPI: ***\n")
    os.system('python -m twine upload dist/* --verbose')  # Requires ~/.pypirc Keys

    print("\n*** Release published successfully. ***\n")

    sys.exit()


setup(
    name='wasd',
    version='1.0.94',
    description='The Kostyan Selenium Wrapper',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Aurococcus/wasd',
    author='Kostyan Opasnost',
    author_email='aurococcus@gmail.com',
    maintainer='Kostyan',
    license="MIT",
    packages=find_packages(),
    install_requires=[
        'pip>=20.0.2',
        'setuptools',
        'wheel',
        'pytest>=5.3.5;python_version>="3"',
        'selenium>=3.141.0',
        'lxml',
        'cssselect',
        'pyyaml',
        'invoke',
        'colorlog',
        'pyhamcrest',
        'termcolor'
    ],
    include_package_data = True,
    entry_points={
        'console_scripts': [
            "wasd = wasd.cli.scaffold:main"
        ],
        'pytest11': ['wasd.fixtures = wasd.core.fixtures']
    }
)



import logging
import colorlog
import shutil
from contextlib import contextmanager
import inspect
from termcolor import colored, cprint
from wasd.core import session


STEP = 60

formatter = colorlog.ColoredFormatter(
    "%(log_color)s[%(levelname).1s] %(message)s",
    datefmt=None,
    reset=True,
    log_colors={
        'DEBUG':    'cyan',
        'INFO':     'green',
        'WARNING':  'yellow',
        'ERROR':    'red',
        'CRITICAL':	'bold_red,bg_white',
        'STEP':		'bold_green'
    },
    secondary_log_colors={},
    style='%'
)

logging.addLevelName(STEP, 'STEP')
handler = colorlog.StreamHandler()
handler.setFormatter(formatter)

LOGGER = colorlog.getLogger(__name__)
LOGGER.addHandler(handler)
LOGGER.setLevel(logging.DEBUG)

#
#
#

__fake_formatter = colorlog.ColoredFormatter(
    "%(message)s",
    reset=True,
    style='%'
)

logging.addLevelName(1001, 'PRINT')
__fake_handler = colorlog.StreamHandler()
__fake_handler.setFormatter(__fake_formatter)

__fake_logger = colorlog.getLogger('PRINT')
__fake_logger.addHandler(__fake_handler)
__fake_logger.setLevel(logging.DEBUG)


def __log_substep(message, limit=None):
    global __fake_logger
    text = colored(_prep_msg(message, 4, limit), 'cyan')
    __fake_logger.log(61, text)


def __log_step(message, limit=None):
    global __fake_logger
    text = colored(_prep_msg(message, 2, limit), 'green', attrs=['bold'])
    __fake_logger.log(61, text)


__c_frame = 0
__p_frame = 0
__c_lvl = 0
__p_lvl = 0


def log_step(message, limit=None):
    if not session.steps:
        return
    global __c_frame, __p_frame, __c_lvl, __p_lvl

    lvl = 0
    while inspect.stack()[lvl].function != "browser":
        if inspect.stack()[lvl].function in ["pytest_pyfunc_call", "_multicall"]:
            lvl -= 1
            break

        lvl += 1

    lvl -= 1
    __c_lvl = inspect.stack()[lvl+1].frame.f_lineno
    __c_frame = id(inspect.stack()[lvl].frame)

    if lvl == 1:
        __log_step(message)
    else:
        human_str = ' '.join([i.capitalize() for i in inspect.stack()[lvl].function.strip('_').split('_')])
        frame = inspect.stack()[lvl].frame
        args, _, _, values = inspect.getargvalues(frame)
        argspec = ', '.join([f"'{values[i]}'" for i in args[1:]])
        if __c_frame != __p_frame:
            __log_step(human_str + " " + argspec)
        else:
            if __c_lvl != __p_lvl:
                __log_step(human_str + " " + argspec)

        __log_substep(message)

    __p_frame = __c_frame
    __p_lvl = __c_lvl


def _prep_msg(message, indent, limit):
    term_width = shutil.get_terminal_size().columns

    msg = message[:limit] if limit else message[:term_width * 3]
    if len(msg) < len(message):
        msg += " ..."
    return ' ' * indent + msg
